from .DataStorage.Alphabet import GreekAlphabet, RomanNumberAlphabet
from .DataType.GreekRomanType import GreekNumber, RomanNumber

class GreekConvert():

    def change_capital(self, capital:bool):
        """Changing the case

        Args:
            capital (bool): Upper or lower register
        """
        self._capital = capital

    def create_greek_number(self, number: str, positional: bool=False) -> GreekNumber:
        """Creating a Greek number

        Args:
            value (str): The value of the number
            positional (bool, optional): Positional or not. Defaults to False.

        Returns:
            GreekNumber: The Greek number
        """
        return GreekNumber(number=number, positional=positional, 
                           capital=self._capital, debug=self._debug)

    def convert(self, number: int) -> str:
        """Converting an Arabic number to a Greek one

        Args:
            number (int): The number to convert

        Returns:
            str: The converted number
        """
        return GreekNumber(number=number, positional=False, 
                           capital=self._capital, debug=self._debug)._value
        
    def convert_position(self, number: int) -> str:
        """Converting an Arabic number to a positional Greek number

        Args:
            number (int): The number to convert

        Returns:
            str: The converted number
        """
        return GreekNumber(number=number, positional=True, 
                           capital=self._capital, debug=self._debug)._value

    def convert_to_arabic(self, numeral: str) -> int:
        """Converting a Greek or Roman number to an Arabic one

        Args:
            numeral (str): The number to convert

        Returns:
            int: The converted number
        """
        return GreekNumber(value=numeral, positional=False, 
                           capital=self._capital, debug=self._debug).get_number()
    
    def covert_to_position_arabic(self, numeral: str) -> int:
        """Converting a positional Greek number to an Arabic one

        Args:
            numeral (str): The number to convert

        Returns:
            int: The converted number
        """
        return GreekNumber(value=numeral, positional=True, 
                           capital=self._capital, debug=self._debug).get_number()

    def __init__(self, capital:bool=False, debug:bool=False):
        """Initializing a class

        Args:
            capital (bool, optional): Upper or lower case. Defaults to False.
        """
        self._debug = debug
        self._capital = capital

    def unicode_to_name(self, greek_numeral: str) -> str:
        """Convert Unicode Greek numeral to name

        Args:
            greek_numeral (str): Greek numeral in Unicode

        Raises:
            ValueError: If invalid symbol is encountered

        Returns:
            str: Name of the Greek numeral
        """
        result = ""
        greek_alphabet = (
            GreekAlphabet.GREEK_ALPHABET_DICT_CAPITAL
            if self._capital
            else GreekAlphabet.GREEK_ALPHABET_DICT
        )
        for char in greek_numeral:
            if char in greek_alphabet:
                result += greek_alphabet[char] + " "
            else:
                raise ValueError(f"Invalid symbol: {char}")
        return result.strip()

    def name_to_unicode(self, name: str) -> str:
        """Convert name of Greek numeral to its Unicode representation

        Args:
            name (str): Name of Greek numeral

        Raises:
            ValueError: If invalid name is encountered

        Returns:
            str: Unicode representation of Greek numeral
        """
        greek_alphabet_dict = GreekAlphabet.GREEK_ALPHABET_DICT
        greek_alphabet_dict_capital = GreekAlphabet.GREEK_ALPHABET_DICT_CAPITAL
        reverse_dict = {v: k for k, v in greek_alphabet_dict.items()}
        reverse_dict_capital = {v: k for k, v in greek_alphabet_dict_capital.items()}
        result = ""
        for word in name.split():
            if word in reverse_dict:
                result += reverse_dict[word]
            elif word in reverse_dict_capital:
                result += reverse_dict_capital[word]
            else:
                raise ValueError(f"Invalid name: {word}")
        return result

class RomanConvert():
    
    def convert(self, arabic_number: int) -> str:
        """Convert Arabic number to Roman numeral

        Args:
            arabic_number (int): Number to convert

        Returns:
            str: Roman numeral representation
        """
        return self._convert_arabic_to_roman(arabic_number)
    
    def create_roman_number(self, arabic_number: int) -> RomanNumber:
        """Create a Roman numeral from an Arabic number

        Args:
            arabic_number (int): The number to convert

        Returns:
            RomanNumber: The Roman numeral representation
        """
        return RomanNumber(arabic_number)
    
    def convert_to_arabic(self, roman_numeral: str) -> int:
        """Convert Roman numeral to Arabic number

        Args:
            roman_numeral (str): Roman numeral to convert

        Returns:
            int: Arabic number representation
        """
        return self._convert_roman_to_arabic(roman_numeral)
    
    def _convert_arabic_to_roman(self, number: int) -> RomanNumber:
        if not (isinstance(number, int)):
            raise TypeError("Число должно быть целым числом и иметь тип int")
        display_numerals = []
        input_num = number
        for numeral, value in RomanNumberAlphabet.ROMAN_NUMERAL_LIST:
            if input_num // value > 0:
                count = input_num // value
                input_num -= count * value
                display_numerals.append(numeral * count)
            else:
                continue
        return ''.join(display_numerals)

    def _convert_roman_to_arabic(self, roman: str) -> int:
        number = 0
        i = 0
        while i < len(roman):
            if i+1 < len(roman) and roman[i:i+2] in RomanNumberAlphabet.ROMAN_NUMERAL_DICT:
                number += RomanNumberAlphabet.ROMAN_NUMERAL_DICT[roman[i:i+2]]
                i += 2
            elif roman[i] in RomanNumberAlphabet.ROMAN_NUMERAL_DICT:
                number += RomanNumberAlphabet.ROMAN_NUMERAL_DICT[roman[i]]
                i += 1
            else:
                raise ValueError(f"Invalid name: {roman}")
        return number


from ..DataStorage.Alphabet import GreekAlphabet, RomanNumberAlphabet

class BaseNumberVirtual():
    _number: int
    _value: list
    _positional: bool
    _capital: bool
    _debug: bool
    _supported_type = (int,)

    def get_number(self) -> int:
        return self._number
    
    def set_number(self, number: int) -> None:
        self._number = number

    def __init__(self, number: int = None, value: str = None, positional: bool = False, capital: bool = False, debug: bool = False) -> None:
        raise NotImplementedError("This is an abstract class")
        self._number = number
        self._value = number
        self._positional = positional
        self._capital = capital
        self._debug = debug

    def _create_instance(self, number: int) -> object:
        return self.__class__(number)
    
    def _update_value(self, number: int) -> None:
        self._number = number

    def __add__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._create_instance(self._number + other._number)
        elif isinstance(other, self._supported_type):
            return self._create_instance(self._number + other)
        else:
            raise TypeError("Unsupported operand type")
    
    def __sub__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._create_instance(self._number - other._number)
        elif isinstance(other, self._supported_type):
            return self._create_instance(self._number - other)
        else:
            raise TypeError("Unsupported operand type")
        
    def __mul__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._create_instance(self._number * other._number)
        elif isinstance(other, self._supported_type):
            return self._create_instance(self._number * other)
        else:
            raise TypeError("Unsupported operand type")
        
    def __truediv__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._create_instance(self._number / other._number)
        elif isinstance(other, self._supported_type):
            return self._create_instance(self._number / other)
        else:
            raise TypeError("Unsupported operand type")
        
    def __floordiv__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._create_instance(self._number // other._number)
        elif isinstance(other, self._supported_type):
            return self._create_instance(self._number // other)
        else:
            raise TypeError("Unsupported operand type")
    
    def __mod__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._create_instance(self._number % other._number)
        elif isinstance(other, self._supported_type):
            return self._create_instance(self._number % other)
        else:
            raise TypeError("Unsupported operand type")
        
    def __pow__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._create_instance(self._number ** other._number)
        elif isinstance(other, self._supported_type):
            return self._create_instance(self._number ** other)
        else:
            raise TypeError("Unsupported operand type")
    
    def __eq__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._number == other._number
        elif isinstance(other, self._supported_type):
            return self._number == other
        else:
            raise TypeError("Unsupported operand type")
    
    def __ne__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._number != other._number
        elif isinstance(other, self._supported_type):
            return self._number != other
        else:
            raise TypeError("Unsupported operand type")
        
    def __lt__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._number < other._number
        elif isinstance(other, self._supported_type):
            return self._number < other
        else:
            raise TypeError("Unsupported operand type")
        
    def __le__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._number <= other._number
        elif isinstance(other, self._supported_type):
            return self._number <= other
        else:
            raise TypeError("Unsupported operand type")
    
    def __gt__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._number > other._number
        elif isinstance(other, self._supported_type):
            return self._number > other
        else:
            raise TypeError("Unsupported operand type")
        
    def __ge__(self, other):
        if isinstance(other, BaseNumberVirtual):
            return self._number >= other._number
        elif isinstance(other, self._supported_type):
            return self._number >= other
        else:
            raise TypeError("Unsupported operand type")
    
    def __iadd__(self, other):
        if isinstance(other, BaseNumberVirtual):
            self._number += other._number
            self._update_value(self._number)
        elif isinstance(other, self._supported_type):
            self._number += other
            self._update_value(self._number)
        else:
            raise TypeError("Unsupported operand type")
        return self
    
    def __isub__(self, other):
        if isinstance(other, BaseNumberVirtual):
            self._number -= other._number
            self._update_value(self._number)
        elif isinstance(other, self._supported_type):
            self._number -= other
            self._update_value(self._number)
        else:
            raise TypeError("Unsupported operand type")
        return self
    
    def __imul__(self, other):
        if isinstance(other, BaseNumberVirtual):
            self._number *= other._number
            self._update_value(self._number)
        elif isinstance(other, self._supported_type):
            self._number *= other
            self._update_value(self._number)
        else:
            raise TypeError("Unsupported operand type")
        return self
    
    def __itruediv__(self, other):
        if isinstance(other, BaseNumberVirtual):
            self._number /= other._number
            self._update_value(self._number)
        elif isinstance(other, self._supported_type):
            self._number /= other
            self._update_value(self._number)
        else:
            raise TypeError("Unsupported operand type")
        return self
    
    def __ifloordiv__(self, other):
        if isinstance(other, BaseNumberVirtual):
            self._number //= other._number
            self._update_value(self._number)
        elif isinstance(other, self._supported_type):
            self._number //= other
            self._update_value(self._number)
        else:
            raise TypeError("Unsupported operand type")
        return self
    
    def __imod__(self, other):
        if isinstance(other, BaseNumberVirtual):
            self._number %= other._number
            self._update_value(self._number)
        elif isinstance(other, self._supported_type):
            self._number %= other
            self._update_value(self._number)
        else:
            raise TypeError("Unsupported operand type")
        return self
    
    def __ipow__(self, other):
        if isinstance(other, BaseNumberVirtual):
            self._number **= other._number
            self._update_value(self._number)
        elif isinstance(other, self._supported_type):
            self._number **= other
            self._update_value(self._number)
        else:
            raise TypeError("Unsupported operand type")
        return self
    
    def __neg__(self):
        return -self._number
    
    def __pos__(self):
        return +self._number

class GreekNumber(BaseNumberVirtual):

    def set_number(self, number: int) -> None:
        self._number = number
        if not self._positional:
            self._convert_arabic_to_greek(number)
        else:
            self._convert_arabic_to_position_greek(number)
    
    def set_positional(self, positional: bool) -> None:
        self._positional = positional
        if positional:
            self._convert_arabic_to_position_greek(self._number)
        else:
            self._convert_arabic_to_greek(self._number)

    def set_capital(self, capital: bool) -> None:
        self._capital = capital
        if not self._positional:
            self._convert_arabic_to_greek(self._number)
        else:
            self._convert_arabic_to_position_greek(self._number)
    
    def get_positional(self) -> bool:
        return self._positional
    
    def get_capital(self) -> bool:
        return self._capital

    def __init__(self, number: int = None, value: str = None, positional: bool = False, capital: bool = False, debug: bool = False) -> None:
        self._capital = capital
        self._debug = debug
        self._positional = positional
        self._number = number
        if number == None and value == None:
            raise ValueError("You must specify a number")
        if not value and not positional:
            self._convert_arabic_to_greek(number)
        elif not value and positional:
            self._convert_arabic_to_position_greek(number)
        elif value and not positional:
            self._convert_greek_to_arabic(value)
        elif value and positional:
            self._convert_position_greek_to_arabic(value)

    def _create_instance(self, number: int) -> object:
        return self.__class__(number, positional=self._positional, capital=self._capital)

    def __iter__(self):
        for item in self._value:
            yield item

    def __str__(self) -> str:
        return f"{''.join(self._value)}"

    def __repr__(self) -> str:
        return f"{''.join(self._value)}"

    def __len__(self) -> int:
        return len(self._value)

    def get_str(self) -> str:
        """Converting a Unicode Greek number to a name

        Raises:
            ValueError: If an invalid character is encountered

        Returns:
            str: The name of the Greek number
        """
        result = ""
        greek_alphabet = (
            GreekAlphabet.GREEK_ALPHABET_DICT_CAPITAL
            if self._capital
            else GreekAlphabet.GREEK_ALPHABET_DICT
        )
        for item in self._value:
            if item in greek_alphabet:
                result += greek_alphabet[item] + " "
            else:
                raise ValueError(f"Invalid character {item}")
        return result.strip()

    def _convert_arabic_to_greek(self, number: int) -> None:
        if not (isinstance(number, int)):
            raise TypeError("The number must be an integer and be of type int")
        greek_numerals_list = (
            GreekAlphabet.GREEK_NUMERAL_LIST_CAPITAL 
            if self._capital 
            else GreekAlphabet.GREEK_NUMERAL_LIST
        )
        display_numerals = []
        hass_pow = False
        power_value = 0
        input_num = number
        while input_num > 0:
            for numeral, _value in reversed(greek_numerals_list):
                power_value = 0
                if len(str(input_num)) > 3:
                    power_value = ((len(str(input_num)) - 1)//3)
                    _value *= 1000**power_value
                    hass_pow = True
                if input_num // _value > 0:
                    if self._debug:
                        print(f"Processing simpol:{numeral}, целое:{number // _value}, остаток:{number % _value}, число:{number}, \
                              значение:{_value}, степень:{(len(str(number)) - 1)//3}")
                    input_num = input_num % _value
                    display_numerals.append(numeral)
                    if power_value:
                        for _ in range(power_value):
                            display_numerals.append("_")
                    if hass_pow:
                        hass_pow = False
                        break
                else:
                    continue
        self._value = ''.join(display_numerals)

    def _convert_arabic_to_position_greek(self, number: int) -> str:
        if not (isinstance(number, int)):
            raise TypeError("The number must be an integer and be of type int")
        greek_numerals_dict = (
            GreekAlphabet.GREEK_NUMERAL_DICT_CAPITAL 
            if self._capital 
            else GreekAlphabet.GREEK_NUMERAL_DICT
        )
        reverse_dict = {v: k for k, v in greek_numerals_dict.items()}
        result = ""
        input_num = number
        derative_remains_list = []
        if self._debug:
            print(f"_convert_arabic_to_position_greek input = {number}")
        while input_num > 0:
            if self._debug:
                print(f"numder: {input_num}, numder % 1000: {input_num % 1000}, numder // 1000: {input_num // 1000}")
            derative_remains_list.append(input_num % 1000)
            input_num = input_num // 1000
        for item in reversed(derative_remains_list):
            if self._debug:
                print(f"result = {result}, i = {item}")
            if len(result) > 0:
                if result[-1] == " ":
                    result = result.strip() + "~"
            for key, _value in reversed(reverse_dict.items()):
                if item // key > 0:
                    result += _value
                    item = item % key
                    if self._debug:
                        print(f"result = {result}, i = {item} in for, key = {key}, _value = {_value}")
            result += " "
        self._value = result.strip()

    def _convert_greek_to_arabic(self, greek_numeral: str) -> int:
        number = 0
        if not (isinstance(greek_numeral, str)):
            raise TypeError("The number must be a string and be of type string")
        power_num = 0
        last_number = 0
        greek_numeral_dict = (
            GreekAlphabet.GREEK_NUMERAL_DICT_CAPITAL 
            if self._capital 
            else GreekAlphabet.GREEK_NUMERAL_DICT
        )
        if self._debug:
            print(f"_convert_greek_to_arabic input = {greek_numeral}")
        for char in greek_numeral:
            if char == "_":
                power_num += 1
                continue
            if self._debug:
                print(f"Processing simpol:{char}, число:{number}, значение:{greek_numeral_dict[char]}, степень:{power_num}")
            if char in greek_numeral_dict:
                last_number *= 1000 ** power_num
                number += last_number
                last_number = greek_numeral_dict[char]
                power_num = 0
            else:
                raise ValueError(f"Invalid symbol: {char}")
        if power_num > 0:
            last_number *= 1000 ** power_num
            power_num = 0
        number += last_number
        self._number = number
    
    def _convert_position_greek_to_arabic(self, greek_numeral: str) -> int:
        if not (isinstance(greek_numeral, str)):
            raise TypeError("The number must be a string and be of type string")
        number = 0
        greek_numeral_dict = (
            GreekAlphabet.GREEK_NUMERAL_DICT_CAPITAL 
            if self._capital 
            else GreekAlphabet.GREEK_NUMERAL_DICT
        )
        if self._debug:
            print(f"_convert_position_greek_to_arabic input = {greek_numeral}")
        for index, item in enumerate(reversed(greek_numeral.split("~"))):
            if self._debug:
                print(f"index: {index}, item: {item}, 1000**{index}")
            for item_sub in item:
                if self._debug:
                    print(f"item_sub: {item_sub}")
                if item_sub in greek_numeral_dict:
                    number += greek_numeral_dict[item_sub] * 1000**index
                else:
                    raise ValueError(f"Invalid symbol: {item_sub}")
        self._number = number


class RomanNumber(BaseNumberVirtual):

    def get_value(self) -> str:
        return ''.join(self._value)

    def __init__(self, number: int) -> None:
        if not number:
            raise ValueError("You must specify a number")
        self._number = number
        self._convert_arabic_to_roman(number)

    def _convert_arabic_to_roman(self, number: int) -> None:
        if not (isinstance(number, int)):
            raise TypeError("The number must be an integer and be of type int")
        display_numerals = []
        input_num = number
        for numeral, _value in RomanNumberAlphabet.ROMAN_NUMERAL_LIST:
            if input_num // _value > 0:
                count = input_num // _value
                input_num -= count * _value
                display_numerals.append(numeral * count)
            else:
                continue
        self._value = display_numerals

    def __iter__(self):
        for item in self._value:
            yield item

    def __getitem__(self, item):
        return self._value[item]

    def __str__(self) -> str:
        return f"{''.join(self._value)}"

    def __repr__(self) -> str:
        return f"{''.join(self._value)}"

    def __len__(self) -> int:
        return len(self._value)


from setuptools import setup, find_packages

def readme():
  with open('README.md', 'r') as f:
    return f.read()

setup(
  name='GreekRomanUtils',
  version='1.0.3',
  author='PKALab',
  author_email='kpalab@pkotlyarov.ru',
  description='The module converts Arabic numerals to Greek and Roman numbers',
  long_description=readme(),
  long_description_content_type='text/markdown',
  url='https://github.com/PavelKotDRM/GreekRomanNumerical',
  packages=find_packages(),
  classifiers=[
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'License :: OSI Approved :: Apache Software License'
  ],
  keywords='python,greek,roman',
  project_urls={
    'Documentation': 'https://github.com/PavelKotDRM/GreekRomanNumerical'
  },
  python_requires='>=3.7'
)